<?php

/**
 * The admin-specific functionality of the plugin.
 *
 * @link       https://emote.com
 * @since      1.0.0
 *
 * @package    Emote
 * @subpackage Emote/admin
 */
class Cms_Import {

  /**
   * The ID of this plugin.
   *
   * @since    1.0.0
   * @access   private
   * @var      string    $plugin_name    The ID of this plugin.
   */
  private $plugin_name;

  /**
   * The version of this plugin.
   *
   * @since    1.0.0
   * @access   private
   * @var      string    $version    The current version of this plugin.
   */
  private $version;

  private $version_number = 'v671';

  /**
   * Initialize the class and set its properties.
   *
   * @since    1.0.0
   * @param      string    $plugin_name       The name of this plugin.
   * @param      string    $version    The version of this plugin.
   */
  public function __construct( $plugin_name, $version ) {

    $this->plugin_name = $plugin_name;
    $this->version = $version;
    $this->ezoic_port = '1337';
  }

  /**
   * Register the stylesheets for the admin area.
   *
   * @since    1.0.0
   */
  public function enqueue_styles() {

    /**
     * This function is provided for demonstration purposes only.
     *
     * An instance of this class should be passed to the run() function
     * defined in Ezoic_Cms_Loader as all of the hooks are defined
     * in that particular class.
     *
     * The Ezoic_Cms_Loader will then create the relationship
     * between the defined hooks and the functions defined in this
     * class.
     */

    wp_enqueue_style( $this->plugin_name, plugin_dir_url( __FILE__ ) . 'css/ezoic-cms-import.css', array(), $this->version, 'all' );

  }

  /**
   * Register the JavaScript for the admin area.
   *
   * @since    1.0.0
   */
  public function enqueue_scripts() {

    /**
     * This function is provided for demonstration purposes only.
     *
     * An instance of this class should be passed to the run() function
     * defined in Ezoic_Cms_Loader as all of the hooks are defined
     * in that particular class.
     *
     * The Ezoic_Cms_Loader will then create the relationship
     * between the defined hooks and the functions defined in this
     * class.
     */

    wp_enqueue_script( $this->plugin_name, plugin_dir_url( __FILE__ ) . 'js/ezoic-cms-import.js', array( 'jquery' ), $this->version, false );

  }

  /**
   * These are the routes.
   * import creates an import event in wp crons
   * initiate calls the scheduled import cron
   * verify searches for the built CSV and zip files and returns their size
   * viewuploadlogs gives the message from the Go server as to whether or not upload was successful
   * clearlogs willclear the upload logs
   */
  public function ez_cms_import_endpoint(){
    register_rest_route( 'ezoic-emote/' . $this->version_number, '/import', array(
      'methods' => WP_REST_Server::READABLE,
      'callback' => array($this, 'ez_cms_parse_table_into_csv_and_send'),
    ) );
    register_rest_route( 'ezoic-emote/' . $this->version_number, '/verify', array(
      'methods' => WP_REST_Server::READABLE,
      'callback' => array($this, 'verify_import_files_built'),
    ) );
    register_rest_route( 'ezoic-emote/' . $this->version_number, '/viewuploadlogs', array(
      'methods' => WP_REST_Server::READABLE,
      'callback' => array($this, 'view_upload_logs'),
    ) );
    register_rest_route( 'ezoic-emote/' . $this->version_number, '/verifyfilename', array(
      'methods' => WP_REST_Server::READABLE,
      'callback' => array($this, 'verify_file_name'),
    ) );
    register_rest_route( 'ezoic-emote/' . $this->version_number, '/verifyuploadevent', array(
      'methods' => WP_REST_Server::READABLE,
      'callback' => array($this, 'verify_upload_event'),
    ) );
    register_rest_route( 'ezoic-emote/' . $this->version_number, '/checkcompletion', array(
      'methods' => WP_REST_Server::READABLE,
      'callback' => array($this, 'check_completion'),
    ) );

  }

  /**
   * This function initiates zipping all files in the temp dir into one file, then names the file
   * according to the document root, then sends the zip file to our production server
   */
  public function ez_cms_file_upload(){
    $temp_dir = get_temp_dir();
    $logs_path = $temp_dir . "file_upload_logs.txt";
    $output_dir = get_temp_dir();
    error_log("have initiated file upload");

    $zipped_result = $this->zip_all_files();
    error_log($zipped_result);

    $file_path = $output_dir . "files.zip";

    $host = $this->find_host();

    $filename = $host . "-dInE96LZv7Jp-commentOnly.zip";

    error_log("Filename is " . $filename . "\n");

    $response = wp_remote_post('http://content-import.ezoic.com/upload', [
      'body' => '--thisisoutrageouslystupid' . "\r\n" . 'Content-Disposition: form-data; name="file"; filename="' .
        $filename . "\"\r\n" . 'Content-Type: application/zip' . "\r\n\r\n" .
	file_get_contents($file_path) . "\r\n--thisisoutrageouslystupid--\r\n",
      'headers' => ['Content-Type' => 'multipart/form-data; boundary=thisisoutrageouslystupid'],
      'method' => 'POST'
    ]);

    error_log(print_r($response, true));

    unlink( $file_path );
    return $response;
  }

  public function zip_all_files(){
    $temp_dir = get_temp_dir();
    $files = array(
      "comments" => $temp_dir . "comments.csv"
    );
    $write_dir = $temp_dir . "files.zip";

    $zip = new ZipArchive();

    $file_opened = $zip->open( $write_dir, ZipArchive::CREATE | ZipArchive::OVERWRITE );
    foreach ( $files as $file_path ){
      $path_array = explode( "/", $file_path );
      $filename = array_pop( $path_array );
      $file_added = $zip->addFile( $file_path, $filename );
      if(!$file_added){
        return false;
      }
    }

    // Zip archive will be created only after closing object
    $zip->close();
    return true;
  }

  public function create_scheduled_event( $data ) {

    $scheduled_event_time = wp_next_scheduled( 'import_event_hook' );
    // schedule cron
    if( $scheduled_event_time ){
      $scheduled_crons = _get_cron_array();
      $already_scheduled_resp_data = array(
        'message' => "we are getting ready to process your current request, please wait. Event scheduled at $scheduled_event_time",
        'body' => $scheduled_crons
      );
      $response = new WP_REST_Response( $already_scheduled_resp_data, 200 );
      return $already_scheduled_resp_data;
    }
    $now = time();
    wp_schedule_single_event( $now, 'import_event_hook');
    $response_data = array(
      'message' => 'import initiated with req to /initiate',
    );
    $response = new WP_REST_Response( $response_data, 200 );
    return $response;
  }

  public function call_scheduled_event ( $data ) {
    $temp_dir = get_temp_dir();

    $data = exec( 'wp cron event run import_event_hook > /dev/null &', $output, $retval );
    $string_output = print_r( $output, true );
    $string_retval = print_r( $retval, true );

    $response = new WP_REST_Response( $data, 200 );
    return $response;
  }

  public function view_logs(){
    $temp_dir = get_temp_dir();
    $file_path = $temp_dir . 'logs.txt';
    $content = file_get_contents( $file_path );
    $response = new WP_REST_Response( $content, 200 );
    return $response;
  }

  public function view_upload_logs(){
    $temp_dir = get_temp_dir();
    $file_path = $temp_dir . "file_upload_logs.txt";
    $content = file_get_contents( $file_path );
    $response = new WP_REST_Response( $content, 200 );
    return $response;
  }

  public function verify_file_name(){
    $temp_dir = get_temp_dir();

    $file_path = $temp_dir . "name_logs.txt";
    $content = file_get_contents( $file_path );
    $response = new WP_REST_Response( $content, 200 );
    return $response;
  }

  public function verify_import_files_built(){
    $temp_dir = get_temp_dir();
    // $file_path = $temp_dir . 'iterms.csv';
    // $content = file_get_contents($file_path);
    $posts_temp = $temp_dir . "posts.csv";
    $posts = filesize($posts_temp);
    $terms_temp = $temp_dir . "terms.csv";
    $terms = filesize($terms_temp);
    $termrels_temp = $temp_dir . "termrels.csv";
    $termrels = filesize($termrels_temp);
    $images_temp = $temp_dir . "images.zip";
    $images = filesize($images_temp);
    $imagerels_temp = $temp_dir . "imagerels.csv";
    $imagerels = filesize($imagerels_temp);
    $imagerels = filesize( $imagerels_temp );
    $comments_temp = $temp_dir . "comments.csv";
    $comments = filesize( $comments_temp );

    $content = array(
      "posts" => "$posts",
      "terms" => "$terms",
      "termrels" => "$termrels",
      "images" => "$images",
      "imagerels" => "$imagerels",
      "comments" => "$comments"
    );
    $response = new WP_REST_Response( $content, 200 );
    return $response;
  }

  public function verify_upload_event(){
    $cron_array = _get_cron_array();
    $response = new WP_REST_Response( $cron_array, 200 );
    return $response;
  }

  private function find_host(){
    if($_SERVER['SERVER_NAME']){
      return $_SERVER['SERVER_NAME'];
    }
    $array = explode( "/", $_SERVER['DOCUMENT_ROOT'] );
    foreach( $array as $dir ){
      if( strpos( $dir, '.com' ) ){
        return $dir;
      }
    }
  }

  public function check_completion(){
    $temp_dir = get_temp_dir();
    $filename = $temp_dir . "import_completion_logs.txt";
    $contents = file_get_contents( $filename );
    return new WP_REST_Response( $contents, 200 );
  }

  public function ez_cms_parse_table_into_csv_and_send(){
    //prevent timeout for long-running script
    set_time_limit(0);
    $temp_dir = get_temp_dir();

    // comments
    $total_comment_rows = $this->number_rows( 'wp_comments' );
    $total_comment_iterations = $this->calculate_iterations( $total_comment_rows, 10000 );
    $this->create_csv_in_chunks( $total_comment_iterations, $total_comment_rows, 'comments.csv', 'wp_comments' );

    // figure out how to get logs from these functions
    $this->ez_cms_file_upload();
    // update completion log
    $temp_dir = get_temp_dir();
    $filename = $temp_dir . "import_completion_logs.txt";
    error_log("completed");
  }

  private function create_csv_in_chunks( $total_iterations, $total_rows, $filename, $tablename ){
    $offset = 0;
    $append = false;
    for( $i = 1; $i <= $total_iterations; $i++ ) {
      if( $i > 1 ){
        $append = true;
      }
      $batch_size = 10000;
      if( $i === $total_iterations ){
        $batch_size = $total_rows - ( $total_iterations * 10000 );
      }
      $file = $this->table_to_csv( $filename, $tablename, $batch_size, $offset, $append );
      $offset += 10000;
    }
  }

  /**
   * This function is responsible for creating an array of all image paths in the system
   * then, chunk by chunk, calling the functions that zip the files
   */
  private function create_image_zip(){
    $batch_size = 10;
    $image_paths = $this->paths_in_filesystem();
    $number_of_images = count( $image_paths );
    $total_iterations = $this->calculate_iterations( $number_of_images, $batch_size );
    // for each: zip file, send it over
    $offset = 0;
    for ( $i = 0; $i < $total_iterations; $i++ ){
      $this->zip_images_filesystem( array_slice( $image_paths, $offset, $batch_size ) );
      $offset += $batch_size;
    }
  }

  /**
   *
   * does the processing to create a csv in the filesystem, then returns the location of that csv in the filesystem
   * @param string $filename
   * @param string $table
   * @param int $total_rows
   * @param int $offset where to start querying in the DB
   */
  private function table_to_csv( $filename, $table, $total_rows, $offset, $append ){
    global $wpdb;
    $temp_dir = get_temp_dir();
    $output_file = $temp_dir . $filename;

    $quantity = 1000;

    $csv_headers = $this->return_headers( $table );

    $total_iterations = $this->calculate_iterations( $total_rows, 1000 );

    for( $i = 0; $i < $total_iterations; $i++ ){
      if ( $append === false ){
        $write_or_append = 'w';
      } else {
        $write_or_append = 'a';
      }

      if( $table === 'wp_comments' ){
        $sql = $this->comment_sql( $offset, $quantity );
        $rows = $wpdb->get_results( $sql, ARRAY_A );
      }
      // opening a stream, closes outside the for loop
      $fp = fopen($output_file, $write_or_append );
      if( $append === false ) {
        if( $table === 'wp_comments' ){
          array_push( $csv_headers, 'path' );
        }
        fputcsv( $fp, $csv_headers );
      }

      foreach ( $rows as $row ) {
        if( $table === 'wp_comments' ){
          $postID = $row['comment_post_ID'];
          array_push( $row, parse_url( get_permalink( $postID ), PHP_URL_PATH ) );
        }
        fputcsv( $fp, $row );
      }
      $offset += 1000;
      $append = true;
    }

    fclose($fp);
    return $output_file;

  }

  /**
   * returns the CSV document headers for all exported tables
   */
  private function return_headers( $table ){
    if( $table === 'wp_posts' ){
      $headers = array( 'ID', 'post_content', 'post_title', 'post_type', 'post_status' );
    } else if ( $table === 'wp_terms' ){
      $headers = array( 'name', 'term_id', 'description', 'taxonomy' );
    } else if ( $table === 'wp_term_relationships' ){
      $headers = array( 'object_id', 'term_id', 'taxonomy' );
    } else if ( $table === 'wp_postmeta' ){
      $headers = array( 'post_id', 'meta_value' );
    } else if ( $table === 'wp_comments' ){
      $headers  = array( 'comment_ID', 'comment_post_ID', 'comment_author', 'comment_author_email', 'comment_author_url', 'comment_author_IP', 'comment_date',
                         'comment_date_gmt', 'comment_content', 'comment_karma', 'comment_approved', 'comment_agent', 'comment_type', 'comment_parent', 'user_id' );
    }
    return $headers;
  }

  private function calculate_iterations( $total_rows, $batch_size ){
    $rounded_iterations = 0;
    if($total_rows >= $batch_size){
      $num_iterations = $total_rows / $batch_size;
      $rounded_iterations = round( $num_iterations, 0, PHP_ROUND_HALF_UP );
      if( $total_rows > ( $rounded_iterations * $batch_size ) ){
        $rounded_iterations += 1;
        return $rounded_iterations;
      }
      return $rounded_iterations;
    }
    return 1;
  }

  /**
   * finds the total number of rows we will be sending over for posts, terms, term relationships and image relationships
   * This later gets divided into chunks to be sent to Ezoic CMS
   */
  private function number_rows( $table ) {
    global $wpdb;
    $total_rows = 0;
    if( $table === 'wp_posts' ){
      $count_sql = $this->count_post_rows();
      $total_rows = $wpdb->get_results($count_sql)[0]->total;
    } elseif( $table === 'wp_terms' ){
      $count_sql = $this->count_term_rows();
      $total_rows = $wpdb->get_results($count_sql)[0]->total;
    } elseif( $table === 'wp_term_relationships' ){
      $count_sql = $this->count_termrel_rows();
      $total_rows = $wpdb->get_results($count_sql)[0]->total;
    } elseif( $table === 'wp_postmeta' ){
      $count_sql = $this->count_imagerel_rows();
      $total_rows = $wpdb->get_results($count_sql)[0]->total;
    } elseif( $table === 'wp_comments' ){
      $count_sql = $this->count_comment_rows();
      $total_rows = $wpdb->get_results($count_sql)[0]->total;
    }
    return $total_rows;
  }

  /** function to return the sql used to count the number of rows in wp_terms */
  private function count_term_rows(){
    $sql = <<<SQL
    select count(*) as total from wp_terms
SQL;
    return $sql;
  }

  /**
   * @param int $offset which row to start on
   * @param int $quantity how many rows to query for
   * @return string SQL query for terms
   */
  private function term_sql( $offset, $quantity ){
    $sql = <<<SQL
    select name, wpt.term_id, description, taxonomy
    from wp_terms as wpt
    inner join wp_term_taxonomy as wptt on wptt.term_id = wpt.term_id
    limit $offset, $quantity
SQL;
    return $sql;
  }

  /** function to return the sql used to count the number of rows in wp_posts */
  private function count_post_rows(){
    $sql = <<<SQL
    select count(*) as total from wp_posts
    where post_type != 'revision' and post_status != 'trash' and post_status != 'auto-draft' and post_status != 'future'
SQL;
    return $sql;
  }

  /**
   * @param int $offset which row to start on
   * @param int $quantity how many rows to query for
   * @return string SQL query for posts
   */
  private function post_sql( $offset, $quantity ){
    $sql = <<<SQL
    select ID, post_content, post_title, post_type, post_status from wp_posts
    where post_type != 'revision' and post_status != 'trash' and post_status != 'auto-draft' and post_status != 'future' and post_type != 'attachment'
    limit $offset, $quantity
SQL;
    return $sql;
  }

  private function count_termrel_rows(){
    $sql = <<<SQL
    select count(*) as total from wp_term_relationships
SQL;
    return $sql;
  }

  private function count_comment_rows(){
    $sql = <<<SQL
    select count(*) as total from wp_comments
SQL;
    return $sql;
  }

  private function comment_sql( $offset, $quantity ){
    $sql = <<<SQL
    select comment_ID, comment_post_ID, comment_author, comment_author_email, comment_author_url, comment_author_IP, comment_date, comment_date_gmt, comment_content, comment_karma, comment_approved, comment_agent, comment_type, comment_parent, user_id
    from wp_comments
    limit $offset, $quantity
SQL;
    return $sql;
  }

  private function count_imagerel_rows(){
    $sql = <<<SQL
    select count(*) as total from wp_postmeta
    where meta_value='_thumbnail_id'
SQL;
    return $sql;
  }

  private function imagerel_sql( $offset, $quantity ) {
    global $wpdb;
    $sql = <<< SQL
    select a.post_id, b.meta_value
    from wp_postmeta as a
    inner join wp_postmeta as b on b.post_id = a.meta_value
    where a.meta_key = '_thumbnail_id' and b.meta_key = '_wp_attached_file'
    limit $offset, $quantity
SQL;
    return $sql;
  }

  /**
   * @param int $offset which row to start on
   * @param int $quantity how many rows to query for
   * @return string SQL query for term-post relationships
   */
  private function termrel_sql( $offset, $quantity){
    $sql = <<< SQL
    select object_id, wpt.term_id, taxonomy
    from wp_term_relationships as wptr
    inner join wp_terms as wpt
    on wpt.term_id = wptr.term_taxonomy_id
    inner join wp_term_taxonomy as wptt
    on wptt.term_taxonomy_id = wptr.term_taxonomy_id
    limit $offset, $quantity
SQL;
    return $sql;
  }

  /**
   * @param array $files the file paths of the images to zip
   */
  private function zip_images_filesystem( $files ){

    $temp_dir = get_temp_dir();
    $write_dir = $temp_dir . 'images.zip';

    $zip = new ZipArchive();
    if( $zip->open($write_dir) === True ){

      foreach ( $files as $array ){
        $file_path = $array[0];
        $relative_path = $array[1];
        $zip->addFile( $file_path, $relative_path );
      }
    } else {
      $zip->open($write_dir, ZipArchive::CREATE | ZipArchive::OVERWRITE);
      foreach ( $files as $array ){
        $file_path = $array[0];
        $relative_path = $array[1];
        $zip->addFile( $file_path, $relative_path );
      }
    }

    // Zip archive will be created only after closing object
    $zip->close();
  }

  /**
   * Recursively determines the file paths for all images in the uploads directory in the filesystem
   * the default storage place for all WP images
   */
  public function paths_in_filesystem(){
    $uploads_dir = $WP_CONTENT_DIR  . '/uploads';
    $files = new RecursiveIteratorIterator(
      new RecursiveDirectoryIterator($uploads_dir),
      RecursiveIteratorIterator::LEAVES_ONLY
    );
    $paths = array();

    foreach( $files as $name => $file ){
      if ( !$file->isDir() ){
        // Get real and relative path for current file
        $file_path = $file->getRealPath();
        $relative_path = substr($file_path, strlen($uploads_dir) + 1);
        // make sure its a full sized image. if it is, add to archive
        if( !preg_match( '/[0-9]{3}x[0-9]{3}/', $relative_path ) ){
          array_push( $paths, array( $file_path, $relative_path ) );
        }
      }
    }
    return $paths;
  }
}
